import contextlib

from orwynn.log.log import Log


def catch_error(error: Exception):
    """Reraise error and catch it by logger."""
    # TODO: Find way to remove this last function from exception stack to not
    #   be displayed

    # try-except block for correct linting even if error isn't actually passed
    # up the stack
    with contextlib.suppress(Exception), Log.catch(reraise=False):
        raise error


catchr = Log.catch(reraise=True)
"""
Shortcut for Log.catch(reraise=True).

Does not accept additional arguments.
"""



import os
from typing import ClassVar

from orwynn.base.error.errors import MalfunctionError
from orwynn.helpers.web import GenericRequest, GenericResponse
from orwynn.log.helpers import catch_error
from orwynn.proxy.boot import BootProxy
from orwynn.utils import validation
from orwynn.utils.scheme import Scheme


class ErrorHandler:
    """
    Handles outcoming errors from the application.
    Method handle(...) should be redefined in subclass in order to work.

    Attributes:
        E:
            Exception or a list of handled Exceptions.
        PROTOCOL:
            Protocol the handler works with.
        IS_ERROR_CATCH_LOGGED:
            Whether the handled errors should be automatically logged by
            Log.catch. Defaults to True.
    """
    E: ClassVar[type[Exception] | None] = None
    PROTOCOL: Scheme = Scheme.HTTP
    IS_ERROR_CATCH_LOGGED: bool = True

    def __init__(self) -> None:
        if self.E is None:
            raise TypeError(
                f"{self.__class__} error class is not set"
            )
        else:
            validation.validate(self.E, Exception)

        validation.validate(self.PROTOCOL, Scheme)

    @classmethod
    def get_handled_exception_class(cls) -> type[Exception]:
        if cls.E is None:
            raise MalfunctionError(
                f"error handler {cls} handled exception shouldn't be None"
            )
        return cls.E

    @property
    def HandledException(self) -> type[Exception]:
        return self.__class__.get_handled_exception_class()

    def _fw_handle_wrapper(
        self,
        request: GenericRequest,
        error: Exception
    ) -> GenericResponse:
        """
        Actual handler called on error occuring.

        Inside it should always propagate the control to self.handle.
        """
        if (
            self.IS_ERROR_CATCH_LOGGED
            # check without an AppMode importing due to circular issues
            and BootProxy.ie().mode.value != "test"
            and not os.getenv(
                "ORWYNN_IS_CATCH_LOGGING_ENABLED_IN_TESTS",
                False
            )
        ):
            catch_error(error)
        return self.handle(request, error)

    def handle(
        self,
        request: GenericRequest,
        error: Exception
    ) -> GenericResponse:
        raise NotImplementedError()



import sys
from typing import TYPE_CHECKING, Any
from orwynn.log.log import Log
from orwynn.log.config import LogConfig
from orwynn.log.handler import LogHandler
from orwynn.proxy.boot import BootProxy

if TYPE_CHECKING:
    from orwynn.app import AppMode


def configure_log(config: LogConfig, *, app_mode_prod: "AppMode") -> None:
    if config.handlers:
        for handler in config.handlers:
            __add_handler(handler, app_mode_prod)

def __add_handler(handler: LogHandler, app_mode_prod: "AppMode") -> int:
    """Adds a new handler.

    Args:
        handler:
            Log handler object to be added.

    Returns:
        Added handler number.
    """
    sink: Any
    if isinstance(handler.sink, str):
        # For apprc-based configuration parse some string literals into
        # actual objects.
        match handler.sink:
            case "$stdout":
                sink = sys.stdout
            case "$stderr":
                sink = sys.stderr
            case _:
                # For all other string cases consider the sink as a file
                # path
                sink = handler.sink
    else:
        # the sink is given directly to the method as a python object,
        # i.e. add it as it is
        sink = handler.sink

    if BootProxy.ie().mode == app_mode_prod:
        if handler.level is None:
            handler.level = "INFO"
        if handler.serialize is None:
            handler.serialize = True
    else:
        if handler.level is None:
            handler.level = "DEBUG"
        if handler.serialize is None:
            handler.serialize = False

    handler_kwargs: dict = handler.kwargs or {}
    kwargs: dict = dict(
        level=handler.level,
        format=handler.format,
        serialize=handler.serialize,
        **handler_kwargs
    )
    if isinstance(handler.sink, str):
        kwargs["rotation"] = handler.rotation  # type: ignore

    return Log.add(
        sink,
        **kwargs
    )
